from contextlib import contextmanager
from collections import namedtuple

from .compat import is_type, type_name
from .config import Section, Compose


Context = namedtuple('Context', 'section key')

Error = namedtuple('Error', 'context message')


def type_repr(t):
    if is_type(t):
        return repr(t)
    else:
        return t.__name__


class ValidationError(TypeError):

    def __init__(self, message, errors):
        message += '\n' + '\n'.join(self._iter_errors(errors))
        super(ValidationError, self).__init__(message)

    def _iter_errors(self, errors):
        for error in errors:
            if error.context.key:
                ctx = '{}[{}]'.format(error.context.section, error.context.key)
            else:
                ctx = error.context.section or '<config>'
            yield ' - {} => {}'.format(ctx, error.message)


class TypeChecker(object):

    def __init__(self, ctx, value, errors):
        self.ctx = ctx
        self.stack = [value]
        self.errors = errors
        self.path = []

    def visit(self, type_):
        value = self.stack[-1]
        if is_type(type_):
            method_name = 'visit_{}'.format(type_name(type_))
            visit_method = getattr(self, method_name, self.not_implemented)
            visit_method(type_, value)
        elif not isinstance(value, type_):
            self.fail(type_, value)

    def not_implemented(self, type_, value):
        raise NotImplementedError('Type check is not implemented for this '
                                  'type: {!r}'.format(type_))

    @contextmanager
    def push(self, value, path_element):
        self.path.append(path_element)
        self.stack.append(value)
        try:
            yield
        finally:
            self.path.pop()
            self.stack.pop()

    def fail(self, type_, value):
        provided = type_repr(type(value))
        expected = type_repr(type_)
        msg = '"{}" instead of "{}"'.format(provided, expected)
        if self.path:
            msg = '{} - {}'.format(''.join(self.path), msg)
        self.errors.append(Error(self.ctx, msg))

    def visit_Union(self, type_, value):
        args = set(type_.__args__)
        if not len(args) == 2 or not type(None) in args:
            raise NotImplementedError('Union types are supported '
                                      'only as Optional type')

        if value is not None:
            arg = (args - {type(None)}).pop()
            self.visit(arg)

    def visit_List(self, type_, value):
        if isinstance(value, list):
            item_type, = type_.__args__
            for i, item in enumerate(value):
                with self.push(item, '[{!r}]'.format(i)):
                    self.visit(item_type)
        else:
            self.fail(type_, value)

    def visit_Dict(self, type_, value):
        if isinstance(value, dict):
            key_type, val_type = type_.__args__
            for key, val in value.items():
                with self.push(key, '[{!r}]'.format(key)):
                    self.visit(key_type)
                with self.push(val, '[{!r}]'.format(key)):
                    self.visit(val_type)
        else:
            self.fail(type_, value)


def validate_type(ctx, value, type_, errors):
    if isinstance(type_, type) or is_type(type_):
        TypeChecker(ctx, value, errors).visit(type_)
    else:
        message = ('"{}" instead of "{}"'
                   .format(type_repr(type(value)), type_repr(type_)))
        errors.append(Error(ctx, message))


def validate_section(obj, value, name, errors):
    assert isinstance(obj, Section), repr(type(obj))
    ctx = Context(name, None)
    validate_type(ctx, value, dict, errors)
    if not errors:
        for key in obj.__keys__.values():
            ctx = Context(name, key.name)
            if key.name not in value:
                errors.append(Error(ctx, 'missing key'))
            else:
                key_value = value[key.name]
                validate_type(ctx, key_value, key.type, errors)


def validate_config(obj, value, variant, sep, errors):
    assert isinstance(obj, Compose), repr(type(obj))
    ctx = Context(None, None)
    validate_type(ctx, value, dict, errors)
    if errors:
        return

    key = sep.join(('compose', variant))
    if key not in value:
        errors.append(Error(Context(key, None), 'missing section'))
        return

    ctx = Context(key, None)
    compose_section = value[key]
    validate_type(ctx, compose_section, dict, errors)
    if errors:
        return

    for section in obj.__sections__.values():
        ctx = Context(key, section.__section_name__)
        if section.__section_name__ not in compose_section:
            errors.append(Error(ctx, 'missing key'))
            continue

        section_variant = compose_section[section.__section_name__]
        validate_type(ctx, section_variant, str, errors)
        if errors:
            continue

        full_section_name = sep.join((section.__section_name__,
                                      section_variant))
        if full_section_name not in value:
            msg = '"{}" not found'.format(full_section_name)
            errors.append(Error(ctx, msg))
        else:
            section_value = value[full_section_name]
            validate_section(section, section_value, full_section_name, errors)


def validate(conf, data, variant, sep):
    errors = []
    validate_config(conf, data, variant, sep, errors)
    return errors



def compose(data, variant, sep):
    entrypoint = sep.join(('compose', variant))
    compose_config = data[entrypoint]
    compose_result = {}
    for key, value in compose_config.items():
        section_name = sep.join((key, value))
        compose_result[key] = data[section_name]
    return compose_result



import sys
import typing


PY37 = sys.version_info[:2] >= (3, 7)


if PY37:
    def is_type(obj):
        return isinstance(obj, typing._Final)
else:
    def is_type(obj):
        return (isinstance(obj, typing.TypingMeta)
                or isinstance(type(obj), typing.TypingMeta))


if PY37:
    def type_name(t):
        if t._name:
            return t._name
        else:
            return t.__origin__._name
else:
    def type_name(t):
        if isinstance(t, typing.TypingMeta):
            return t.__name__
        else:
            return type(t).__name__.lstrip('_')



import typing  # noqa


class Key(object):

    def __init__(self, name, type_):
        self.name = name
        self.type = type_


class NotInitialized(object):

    def __get__(self, instance, owner):
        raise RuntimeError('Config is not initialized')


class SectionBase(object):
    __keys__ = {}  # type: typing.Dict[str, Key]
    __descriptors__ = {}

    def __init__(self, name):
        self.__section_name__ = name


class SectionMeta(type):

    def __new__(mcs, name, bases, params):
        keys = {key: val for key, val in params.items()
                if isinstance(val, Key)}

        annotations = params.get('__annotations__', {})
        for name, type_ in annotations.items():
            if name not in params:
                keys.setdefault(name, Key(name, type_))

        descriptors = {key: val for key, val in params.items()
                       if getattr(val, '__get__', None) is not None}
        cls = super(SectionMeta, mcs).__new__(mcs, name, bases, params)
        cls.__keys__ = dict(cls.__keys__, **keys)
        cls.__descriptors__ = dict(cls.__descriptors__, **descriptors)

        not_initialized = NotInitialized()
        for attr_name in keys:
            setattr(cls, attr_name, not_initialized)

        return cls


class Section(SectionBase, metaclass=SectionMeta):
    pass


class SectionValue(object):

    def __init__(self, section, value):
        for attr_name, key in section.__keys__.items():
            setattr(self, attr_name, value[key.name])
        self.__section = section
        self.__descriptors = section.__descriptors__

    def __getattr__(self, name):
        if name in self.__descriptors:
            value = self.__descriptors[name].__get__(self, self.__class__)
        else:
            value = getattr(self.__section, name)
        # caching
        setattr(self, name, value)
        return value


class ComposeBase(object):
    __sections__ = {}  # type: typing.Dict[str, Section]
    __initialized = False

    def __init_sections__(self, values):
        if self.__initialized:
            raise RuntimeError('Config is already initialized')
        for attr_name, section in self.__sections__.items():
            value = values[section.__section_name__]
            setattr(self, attr_name, SectionValue(section, value))
        self.__initialized = True


class ComposeMeta(type):

    def __new__(mcs, name, bases, params):
        sections = {key: val for key, val in params.items()
                    if isinstance(val, Section)}

        annotations = params.get('__annotations__', {})
        for name, section_type in annotations.items():
            if name not in params:
                sections.setdefault(name, section_type(name))

        cls = super(ComposeMeta, mcs).__new__(mcs, name, bases, params)
        cls.__sections__ = dict(cls.__sections__, **sections)
        return cls


class Compose(ComposeBase, metaclass=ComposeMeta):
    pass


class key_property(object):

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, instance, owner):
        if instance is None:
            return self
        value = instance.__dict__[self.func.__name__] = self.func(instance)
        return value



from .compose import compose
from .checker import validate, ValidationError


def init(conf, data, variant, sep='.'):
    errors = validate(conf, data, variant, sep)
    if errors:
        raise ValidationError('Configuration is not valid', errors)
    else:
        composed_data = compose(data, variant, sep)
        conf.__init_sections__(composed_data)



from __future__ import absolute_import

import codecs

import toml

from . import data


def init(conf, file_names, variant, sep='_'):
    conf_data = {}
    for file_name in file_names:
        with codecs.open(file_name, encoding='utf-8') as f:
            conf_data.update(toml.loads(f.read()))
    data.init(conf, conf_data, variant, sep=sep)



from setuptools import setup, find_packages

setup(
    name='strictconf',
    version='0.3.0rc2',
    description='Configuration library with type-checking',
    author='Vladimir Magamedov',
    author_email='vladimir@magamedov.com',
    url='https://github.com/vmagamedov/strictconf',
    packages=find_packages(),
    license='BSD',
)
